import click

def present_cases(cases):
    if len(cases) == 0:
        click.secho(" NO TESTS ", bg="yellow", fg="black")
        return -1
    
    fails = []
    for case in cases:
        name = case["name"]
        error = case["error"]

        if error == None:
            click.secho("  ✓ ", fg="green", nl=False)
            click.secho(name, bold=True)
            continue

        if error == "invalid_case":
            click.secho("  ⚠ ", fg="red", nl=False)
            click.secho(name, bold=True, nl=False)
            click.echo(" (invalid testcase)")
        elif error == "runtime":
            click.secho("  ⚠ ", fg="red", nl=False)
            click.secho(name, bold=True, nl=False)
            click.echo(" (runtime error)")
        elif error == "presentation":
            click.secho("  × ", fg="yellow", nl=False)
            click.secho(name, bold=True)
        elif error == "wrong_answer":
            click.secho("  × ", fg="red", nl=False)
            click.secho(name, bold=True)
        
        fails.append(case)
    
    if len(fails) == 0:
        click.secho(" PASS ", bg="green", fg="black")
    else:
        click.secho(" FAIL ", bg="red", fg="black")
        for fail in fails:
            name = fail["name"]
            error = fail["error"]

            click.echo()
            if error == "presentation":
                click.secho("● ", fg="yellow", nl=False)
                click.secho(f"{name}: presentation error")
            elif error == "wrong_answer":
                diff = fail["diff"]

                click.secho("● ", fg="red", nl=False)
                click.secho(f"{name}: wrong answer")
                for line in diff:
                    color = "red" if line.startswith("-") else "green" if line.startswith("+") else None
                    click.secho(f"{line}", fg=color)
            elif error == "runtime":
                stderr = fail["stderr"]

                click.secho("● ", fg="red", nl=False)
                click.secho(f"{name}: runtime error")
                click.echo(stderr)
            elif error == "invalid_case":
                click.secho(f"{name}: invalid test case", fg="red")

def present_ci_cases(cases):
    if len(cases) == 0:
        click.echo("      NO TESTS")
        return -1
    
    fails = []
    for case in cases:
        name = case["name"]
        error = case["error"]

        if error == None:
            click.echo(f"      ✓ {name}")
            continue

        if error == "invalid_case":
            click.echo(f"      ⚠ {name} (invalid testcase)")
        elif error == "runtime":
            click.secho(f"      ⚠ {name} (runtime error)")
        elif error == "presentation":
            click.secho(f"      × {name}")
        elif error == "wrong_answer":
            click.secho(f"      × {name}")
        
        fails.append(case)



import tomllib
import difflib
import os
import subprocess

def run(bin_path, testcases_path):
    with open(testcases_path, "rb") as f:
        cases = tomllib.load(f).items()
        runs = []

        for name, case in cases:
            if "in" not in case or "out" not in case:
                runs.append({
                    "name": name,
                    "error": "invalid_case",
                })
                continue

            result = subprocess.run(
                bin_path,
                input=case.get("in"),
                capture_output=True,
                text=True,
                env=dict(os.environ, LIBC_FATAL_STDERR_="1")
            )

            if result.returncode != 0:
                runs.append({
                    "name": name,
                    "error": "runtime",
                    "stderr": result.stderr,
                })
                continue

            actual = result.stdout
            expected = case.get("out")
            if expected == actual:
                runs.append({
                    "name": name,
                    "error": None,
                })
            else:
                diff = difflib.unified_diff(expected.splitlines(), actual.splitlines(), fromfile="expected", tofile="received", lineterm="")
                line = next(diff, None)
                if line is None:
                    runs.append({
                        "name": name,
                        "error": "presentation",
                    })
                else:
                    fail_diff = [line]
                    for l in diff:
                        fail_diff.append(l)

                    runs.append({
                        "name": name,
                        "error": "wrong_answer",
                        "diff": fail_diff
                    })
        
        return runs



import click
import os
import pyperclip
import subprocess
from .common.exception import NotFoundException
from .common.paths import ensure_cwd, ensure_paths, icanc_path
from .common.rc import config
from .tools.preprocessor import preprocess

@click.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("-s", "--solution", "solution_src", default="solution", help="Which solution file to submit.")
@click.option("-e", "--edit", "open_editor", is_flag=True, help="Open submission file on text editor.")
@click.option("-c", "--copy", is_flag=True, help="Copy submission to clipboard.")
@click.option("-u", "--uglify", is_flag=True, help="Remove all comments.")
def submit(**kwargs):
    """Bundle solution into a single source file for submission."""
    handle_submit(**kwargs)

def handle_submit(judge, problem, solution_src, open_editor, copy, uglify):
    ensure_cwd()
    ensure_paths()
    
    solution_filename = f"{solution_src}.c"
    solution_path_rel = f"./{judge}/{problem}/{solution_filename}"
    solution_path = icanc_path("problems", judge, problem, solution_filename)
    if not os.path.exists(solution_path):
        raise NotFoundException("solution", solution_path_rel)
    submission = preprocess(judge, problem, solution_src, solution_path, {}, uglify)

    submission_dir = icanc_path("submissions", judge, problem)
    submission_path = icanc_path("submissions", judge, problem, solution_filename)
    os.makedirs(submission_dir, exist_ok=True)

    with open(submission_path, "w") as f:
        f.write(submission)

    if open_editor:
        subprocess.run([config["editor"], submission_path])

    if copy:
        pyperclip.copy(submission)



class IcancException(Exception):
    def __init__(self, error, message, hint=None):
        super().__init__(error, message)
        self.error = error
        self.message = message
        self.hint = hint

class FoundException(IcancException):
    def __init__(self, entity, id, hint=None):
        super().__init__(
            f"{entity} found",
            f"{id} already exists.",
            hint,
        )

class NotFoundException(IcancException):
    def __init__(self, entity, id, hint=None):
        super().__init__(
            f"{entity} not found",
            f"{id} does not exist.",
            hint,
        )

class UnknownPragmaException(IcancException):
    def __init__(self, pragma, id, hint=None):
        super().__init__(
            f"unknown pragma {pragma}",
            f"unknown pragma at {id}.",
            hint,
        )

class UnknownIncludeException(IcancException):
    def __init__(self, include, id, hint=None):
        super().__init__(
            "include not found",
            f"{include} not found at {id}",
            hint,
        )

class InvalidSourceException(IcancException):
    def __init__(self, entity, id, hint=None):
        super().__init__(
            f"invalid {entity}",
            f"invalid {entity} at {id}",
            hint,
        )

class InvalidCommandException(IcancException):
    def __init__(self, error, hint=None):
        super().__init__(
            "invalid command",
            error,
            hint
        )

class NetworkException(IcancException):
    def __init__(self, status, error, hint=None):
        super().__init__(
            f"network error ({status})",
            error,
            hint
        )



import os
from .exception import NotFoundException
from importlib.resources import files
from .rc import preload_rc

cwd = os.getcwd()
dirs = [
    "include",
    "problems",
    "templates",
    "submissions",
    "binaries",
]
paths = [os.path.join(cwd, dir) for dir in dirs]
icc_dirs = dict(zip(dirs, paths))

def ensure_paths():
    for dir in paths:
        os.makedirs(dir, exist_ok=True)

def icanc_path(dir, *argv):
    path = icc_dirs.get(dir)
    
    for arg in argv:
        path = os.path.join(path, arg)

    return path

def data_path(path):
    return files("icanc.data").joinpath(path)

def ensure_cwd():
    if not os.path.exists(os.path.join(cwd, "icancrc.toml")):
        raise NotFoundException("config", "icancrc.toml", "Are you running from the project root?")
    
    preload_rc()



import click
from ..common.exception import InvalidSourceException, UnknownIncludeException, UnknownPragmaException
from ..common.paths import icanc_path
from ..common.rc import config
import os
import re

def preprocess(judge, problem, solution, path, cache, uglify):
    author = config.get("author")
    email = config.get("email")
    repository = config.get("repository")

    credits = ""
    if author:
        credits += f" Solution by {author}"
        if email:
            credits += f" <{email}>"
        credits += "\n"
    if repository:
        credits += f" * {repository}\n"
    if author or email or repository:
        credits += " *"

    header = f"""/*
 * {judge}/{problem} {solution}.c
 *{credits}
 * Powered by icanc
 * https://github.com/voxelstack/icanc
 */
"""
    
    return "{}\n{}".format(header, "".join(preprocess_source(path, cache, [], uglify)["source"]))

def preprocess_source(path, cache, included, uglify):
    if path in cache.values():
        return cache[path]
    
    # TODO Relative paths would be clearer for errors.
    filename = f"{os.path.basename(path)}"
    with open(path, "r", encoding="utf-8") as f:
        source = iter(f.readlines())
        preprocessed_source = []
        pragma_once = False
        last_line_empty = False

        lineno = 0
        line = next(source, None)
        while line is not None:
            trimmed = line.lstrip()

            if uglify:
                line = re.sub(r"\s*//.*$", "", line)
                if trimmed.startswith("/*"):
                    while line is not None and "*/" not in line:
                        line = next(source, None)
                        trimmed = line.lstrip()
                        lineno += 1
                    line = re.sub(r"^.*\*/", "", line)
                    continue
                elif re.match(r"^\s*\n", line):
                    line = next(source, None)
                    trimmed = line.lstrip()
                    lineno += 1
                    continue

            # Yeah yeah.
            # String matching is enough here but I do speak AST.
            #
            # https://github.com/voxelstack/atlas/blob/main/src/atlas/comms_derive/src/shareable.rs
            # https://xkcd.com/1171/
            if trimmed.startswith("#pragma"):
                pragma = trimmed[len("#pragma"):].lstrip()
                if pragma.startswith("once"):
                    pragma_once = True
                elif pragma.startswith("icanc"):
                    pragma_icanc = pragma[len("icanc"):].lstrip()
                    if pragma_icanc.startswith("include"):
                            while True:
                                line = next(source, None).lstrip()
                                lineno += 1
                                if line is None or not line.startswith("#include"):
                                    if line.startswith("#pragma icanc end"):
                                        break
                                    raise InvalidSourceException("include directive", f"{filename}:{lineno}", f"Expected #include or #pragma icanc end, got \"{line.strip()}\".")

                                include_file = line[len("#include"):].strip()[1:-1]
                                include_path = icanc_path("include", include_file)
                                if not os.path.exists(include_path):
                                    raise UnknownIncludeException(include_file, f"{filename}:{lineno}")
                                include = preprocess_source(include_path, cache, included, uglify)
                                if not include["once"] or include_path not in included:
                                    if not uglify:
                                        preprocessed_source.append("// BEGIN {}\n".format(include_file))
                                        preprocessed_source.extend(include["source"])
                                        preprocessed_source.append("// END {}\n".format(include_file))
                                    else:
                                        preprocessed_source.extend(include["source"])
                                    
                                    included.append(include_path)
                    else:
                        raise UnknownPragmaException(f"\"{pragma_icanc.strip()}\"", f"{filename}:{lineno}")
                line = next(source, None)
                lineno += 1
                continue

            current_line_empty = len(trimmed) == 0
            # Since we are omitting some lines when generating the submission,
            # it's common to have double line breaks which look bad.
            # Skip them.
            if not (last_line_empty and current_line_empty):
                preprocessed_source.append(line)
            last_line_empty = current_line_empty

            line = next(source, None)
            lineno += 1

    source = {
        "source": preprocessed_source,
        "once": pragma_once,
    }
    cache[path] = source
    return source



import click
import os
import shutil
import subprocess
from .common.exception import FoundException
from .common.paths import data_path

@click.command()
@click.option("--name", prompt=True, default="leetcode")
@click.option("--author", prompt=True)
@click.option("--email", prompt=True)
@click.option("--repository", prompt=True)
@click.option("--git", prompt="Initialize git repository", default=True)
def init(**kwargs):
    """Initialize an icanc project."""
    handle_init(**kwargs)

def handle_init(name, author, email, repository, git):
    dir = os.path.join(os.getcwd(), name)
    if os.path.exists(dir):
        raise FoundException("project", f"./{name}/")
    os.makedirs(dir)

    shutil.copy2(data_path("icancrc.toml"), dir)
    shutil.copy2(data_path("LICENSE"), dir)
    shutil.copytree(data_path("include"), os.path.join(dir, "include"))
    shutil.copytree(data_path("templates"), os.path.join(dir, "templates"))
    shutil.copytree(data_path("problems"), os.path.join(dir, "problems"))
    with open(data_path("README.md"), "r") as src:
        readme = src.read().format(name=name)
        with open(os.path.join(dir, "README.md"), "w") as dst:
            dst.write(readme)

    with open(data_path("icancrc.toml"), "r") as src:
        rc = src.read().format(
            author=f"author=\"{author}\"\n" if author else "",
            email=f"email=\"{email}\"\n" if email else "",
            repository=f"repository=\"{repository}\"\n" if repository else ""
        )
        with open(os.path.join(dir, "icancrc.toml"), "w") as dst:
            dst.write(rc)
    
    if git:
        shutil.copy2(data_path(".gitignore"), dir)
        subprocess.run(["git", "init", dir])
    
    click.secho("\n DONE ", bg="green", nl=False);
    click.secho(f" Your project was created at {name}/", fg="green")
    click.secho("Get started with",  nl=False)
    click.secho(f" cd {name} && icanc --help", bold=True)

def read_git_config(config):
    res = subprocess.run(["git", "config", config], stdout=subprocess.PIPE)
    if res.returncode == 0:
        return res.stdout.rstrip().decode()
    return ""



import click
import os
import subprocess
from ..common.rc import config

def build(src_path, submission_path, bin_path):
    result = subprocess.run(
        [
            config["compiler"],
            submission_path,
            "-o", bin_path,
            "-I", os.path.join(os.getcwd(), "include")
        ],
        capture_output=True,
        text=True
    )

    if result.returncode != 0:
        click.secho(" Build failed ", bg="red", fg="black", err=True)
        # The compiler errors will point to the submission.
        # Print the actual source so you can navigate to it.
        click.echo(src_path)
        click.echo()
        click.echo(result.stderr, err=True)
    
    return result.returncode



import click
import json
import requests
import time
import tomli_w
from ..common.exception import NetworkException, NotFoundException
from ..common.rc import config
from ..tools.auth import get_auth

udebug_url = "https://www.udebug.com"

def fetch_testcases(judge, problem):
    auth = get_auth("icanc.udebug")

    user = auth["user"]
    password = auth["password"]
    if user is None or password is None:
        raise NotFoundException("config", "udebug auth", "Did you authenticate with icanc udebug auth?")
    
    auth = requests.auth.HTTPBasicAuth(user, password)

    src = {}
    click.echo("Downloading udebug testcases...")
    with click.progressbar(list_testcases(judge, problem, auth)) as testcases:
        time.sleep(0.2)
        for testcase in testcases:
            id = testcase["id"]
            user = testcase["user"]

            i = get_input(id, auth)
            time.sleep(0.2)
            o = get_output(id, auth)
            time.sleep(0.2)

            src[f"{id} by {user}"] = {"in": i, "out": o}

    return tomli_w.dumps(src)

def list_testcases(judge, problem, auth):
    udebug_config = config.get("udebug", {})
    min_votes = udebug_config.get("min_votes", 5)
    judge_aliases = udebug_config.get("judge_alias", {})
    judge_alias = judge_aliases.get(judge, judge)

    res = make_request(
        "input_api/input_list/retrieve.json",
        {"judge_alias":  judge_alias, "problem_id": problem},
        auth
    )

    return [r for r in res if int(r["Votes"]) >= min_votes]

def get_input(id, auth):
    res = make_request(
        "input_api/input/retrieve.json",
        {"input_id":  id},
        auth
    )
    
    return res[0]

def get_output(id, auth):
    res = make_request(
        "output_api/output/retrieve.json",
        {"input_id":  id},
        auth
    )

    return res[0]

def make_request(path, params, auth):
    res = requests.get(
        f"{udebug_url}/{path}",
        headers={"accept": "application/json"},
        params=params,
        auth=auth
    )
    if not res.ok:
        raise NetworkException(res.status_code, f"failed to fetch {path}", "Are your credentials correct, and do you have API access to udebug?" if res.status_code == 401 else None)
    
    return json.loads(res.content)



import click
import os
from pathlib import Path
from .submit import handle_submit
import sys
from .tools.builder import build
from .common.paths import ensure_cwd, ensure_paths, icanc_path
from .tools.preprocessor import preprocess
from .tools.runner import run
from .tools.writer import present_ci_cases

@click.command()
@click.option("-j", "--judge", type=str, help="Only test one judge.")
def ci(**kwargs):
    """Test everything with machine friendly output and exit codes."""
    handle_ci(**kwargs)

def handle_ci(judge):
    ensure_cwd()
    ensure_paths()

    problems_dir = icanc_path("problems")
    preprocess_cache = {}

    judges = []
    if judge:
        judges.append(judge)
    else:
        judges = os.listdir(problems_dir)
    
    failed = 0
    for judge in judges:
        failed_judge = 0
        judge_dir = icanc_path("problems", judge)
        problems = os.listdir(judge_dir)
        
        for problem in problems:
            failed_problem = 0
            problem_dir = icanc_path("problems", judge, problem)
            solutions = [s for s in os.listdir(problem_dir) if s.endswith(".c")]

            click.echo(f"{judge}/{problem}")
            
            for solution in solutions:
                failed_solution = 0
                click.secho(f"  ./problems/{judge}/{problem}/{solution}")
                testcases = [s for s in os.listdir(problem_dir) if s.endswith(".toml")]

                solution_path = icanc_path("problems", judge, problem, solution)
                submission_path = icanc_path("submissions", judge, problem, solution)
                binary_dir = icanc_path("binaries", judge, problem)
                binary_path = icanc_path("binaries", judge, Path(solution).stem)
                os.makedirs(binary_dir, exist_ok=True)

                submission_dir = icanc_path("submissions", judge, problem)
                submission_path = icanc_path("submissions", judge, problem, solution)
                os.makedirs(submission_dir, exist_ok=True)

                submission = preprocess(judge, problem, solution, solution_path, preprocess_cache, True)
                with open(submission_path, "w") as f:
                    f.write(submission)

                result = build(solution_path, submission_path, binary_path)
                if result != 0:
                    return -1

                for testcase in testcases:
                    click.echo(f"    < ./problems/{judge}/{problem}/{testcase}")
                    
                    testcase_path = icanc_path("problems", judge, problem, testcase)
                    runs = run(binary_path, testcase_path)
                    failed_solution += sum(1 if r["error"] is not None else 0 for r in runs)
                    present_ci_cases(runs)
                failed_problem += failed_solution
            failed_judge += failed_problem
        failed += failed_judge
    
    if failed == 0:
        click.echo("PASS")
    sys.exit(failed)



import click
import os
import shutil
import subprocess
from .common.exception import FoundException, InvalidCommandException, NotFoundException
from .common.paths import ensure_cwd, ensure_paths, icanc_path
from .common.rc import config
from .tools.udebug import fetch_testcases

@click.group()
def create():
    """Create solutions or testcases."""
    pass

@create.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("-t", "--template", type=str, default="main")
@click.option("-s", "--solution", "solution_dst", default="solution", help="Name for the solution file.")
@click.option("-e", "--edit", "open_editor", is_flag=True, help="Open solution file on text editor.")
def solution(**kwargs):
    create_solution(**kwargs)

def create_solution(judge, problem, template, solution_dst, open_editor):
    ensure_cwd()
    ensure_paths()

    template_filename = f"{template}.c"
    template_path = icanc_path("templates", template_filename)
    if not os.path.exists(template_path):
        raise NotFoundException("template", f"./templates/{template_filename}")
    
    dir = icanc_path("problems", judge, problem)
    os.makedirs(dir, exist_ok=True)

    solution_filename = f"{solution_dst}.c"
    solution_path_rel = f"./problems/{judge}/{problem}/{solution_filename}"
    solution_path = icanc_path("problems", judge, problem, solution_filename)
    if os.path.exists(solution_path):
        raise FoundException("solution", solution_path_rel, "To create multiple solutions set the solution name with the --solution option.")
    shutil.copy2(template_path, solution_path)
    
    click.echo(f"Created blank solution: {solution_path_rel}")

    if open_editor:
        subprocess.run([config["editor"], solution_path])

@create.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("-t", "--testcases", "testcases_dst", default="testcases", help="Name for the testcases file.")
@click.option("-e", "--edit", "open_editor", is_flag=True, help="Open testcases file on text editor.")
@click.option("-u", "--udebug", "udebug", is_flag=True, help="Download testcases from udebug.")
def testcases(**kwargs):
    create_testcases(**kwargs)

def create_testcases(judge, problem, testcases_dst, open_editor, udebug):
    if (not udebug and testcases_dst == "udebug"):
        raise InvalidCommandException("udebug.toml is reserved for udebug testcases", "Change the value of --testcases to something else.")
    
    testcases_set = click.get_current_context().get_parameter_source("testcases_dst") != click.core.ParameterSource.DEFAULT
    if (udebug and testcases_set and testcases_dst != "udebug"):
        raise InvalidCommandException("udebug testcases must be called udebug.toml", "You can leave --testcases unset when creating udebug testcases.")

    ensure_cwd()
    ensure_paths()
    
    dir = icanc_path("problems", judge, problem)
    if not os.path.exists(dir):
        raise NotFoundException("problem", f"{judge}/{problem}")

    if udebug:
        testcases_path = create_udebug_testcases(judge, problem)
    else:
        testcases_path = create_blank_testcases(judge, problem, testcases_dst)

    if open_editor:
        subprocess.run([config["editor"], testcases_path])

def create_blank_testcases(judge, problem, testcases_dst):
    testcases_filename = f"{testcases_dst}.toml"
    testcases_path_rel = f"./problems/{judge}/{problem}/{testcases_filename}"
    testcases_path = icanc_path("problems", judge, problem, testcases_filename)
    if os.path.exists(testcases_path):
        raise FoundException("testcases", testcases_path_rel, "To create multiple testcases set the solution name with the --testcases option.")

    with open(testcases_path, "w", encoding="utf-8") as f:
        f.writelines([
            "# {}/{}\n\n".format(judge, problem),
            "[case1]\n",
            "in = \"\"\"\"\"\"\n",
            "out = \"\"\"\"\"\"\n"
        ])

    click.echo(f"Created blank testcases: {testcases_path_rel}")

    return testcases_path

def create_udebug_testcases(judge, problem):
    testcases_filename = f"udebug.toml"
    testcases_path_rel = f"./problems/{judge}/{problem}/{testcases_filename}"
    testcases_path = icanc_path("problems", judge, problem, testcases_filename)

    with open(testcases_path, "w") as f:
        header = f"# {judge}/{problem}\n# udebug.com\n\n"
        testcases = fetch_testcases(judge, problem)
        f.write(header + testcases)

    click.echo(f"Downloaded udebug testcases: {testcases_path_rel}")
    click.echo(f"\nRemember to vote on useful inputs!")
    click.echo(f"https://www.udebug.com/?search_string={problem}&search_category=0")

    return testcases_path

@click.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("-t", "--template", type=str, default="main")
@click.option("--solution", "solution_dst", default="solution", help="Name for the solution file.")
@click.option("--testcases", "testcases_dst", default="testcases", help="Name for the testcases file.")
@click.option("-e", "--edit", "open_editor", is_flag=True, help="Open solution file on text editor.")
def scaffold(judge, problem, template, solution_dst, testcases_dst, open_editor):
    """Create a solution and testcase files."""
    
    click.echo("Scaffolding {}/{}\n.".format(judge, problem))
    create_solution(judge, problem, template, solution_dst, open_editor)
    create_testcases(judge, problem, testcases_dst, False, False)



import click
import os
from .submit import handle_submit
import time
from .tools.builder import build
from .common.paths import ensure_cwd, ensure_paths, icanc_path
from .common.exception import NotFoundException
from .tools.runner import run
from .tools.writer import present_cases
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

@click.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("-s", "--solution", "solution_src", default="solution", help="Which solution to test.")
@click.option("-t", "--testcases", "testcases_src", multiple=True, help="Which testcases to test against.")
@click.option("-w", "--watch", is_flag=True, help="Rerun on changes to the problem or include directories.")
def test(**kwargs):
    """Test a solution against given testcases."""
    handle_test(**kwargs)

def handle_test(judge, problem, solution_src, testcases_src, watch):
    ensure_cwd()
    ensure_paths()
    
    solution_dir = icanc_path("problems", judge, problem)
    solution_filename = f"{solution_src}.c"
    solution_path = icanc_path("problems", judge, problem, solution_filename)
    submission_path = icanc_path("submissions", judge, problem, solution_filename)
    include_dir = icanc_path("include")
    binary_dir = icanc_path("binaries", judge, problem)
    binary_path = icanc_path("binaries", judge, problem, solution_src)
    testcases_dir = icanc_path("problems", judge, problem)
    os.makedirs(binary_dir, exist_ok=True)

    if not os.path.exists(solution_path):
        raise NotFoundException("solution", f"./problems/{judge}/{problem}/{solution_filename}")

    def clear():
        if os.name == "nt":
            os.system("cls")
        else:
            os.system("clear")

    def test():
        # TODO Only rebuild the list of testcases when a file is created or deleted.
        testcases = []
        testcase_files = []
        if testcases_src:
            testcase_files.extend([f"{t}.toml" for t in testcases_src])
        else:
            files = os.listdir(testcases_dir)
            testcase_files.extend([f for f in files if f.endswith(".toml")])

        for testcase in testcase_files:
            path = icanc_path("problems", judge, problem, testcase)
            if not os.path.exists(path):
                raise NotFoundException("testcases", f"./problems/{judge}/{problem}/{testcase}")
            testcases.append({
                "name": testcase,
                "path": path
            })

        clear()
        if watch:
            click.secho(" WATCH ", bg="cyan", fg="white", nl=False)
        else:
            click.secho(" RUN ", bg="cyan", fg="white", nl=False)
        click.echo(f" ./problems/{judge}/{problem}/", nl=False)
        click.secho(solution_filename, bold=True)

        for testcase in testcases:
            click.echo(f"\n< ./problems/{judge}/{problem}/{testcase['name']}")

            handle_submit(judge, problem, solution_src, False, False, False)
            
            result = build(solution_path, submission_path, binary_path)
            if result != 0:
                return

            runs = run(binary_path, testcase["path"])
            present_cases(runs)

        if watch:
            click.echo("\nWatching for changes. ^C to stop.")

    class FileWatcher(FileSystemEventHandler):
        def __init__(self):
            FileSystemEventHandler.__init__(self)

        def on_modified(self, event):
            clear()
            test()

    if watch:
        watcher = FileWatcher()
        observer = Observer()
        observer.schedule(watcher, path=solution_dir, recursive=True)
        observer.schedule(watcher, path=include_dir, recursive=True)
        observer.start()

        try:
            test()
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
    else:
        test()



import click
from .tools.auth import set_auth, get_auth

@click.group()
def udebug():
    """udebug integration."""
    pass

@udebug.command()
@click.option("--user", prompt=True)
@click.option("--password", prompt=True, hide_input=True)
def auth(**kwargs):
    """Authenticate with udebug."""
    handle_auth(**kwargs)

def handle_auth(user, password):
    set_auth("icanc.udebug", user, password)
    click.echo("Your credentials were stored!")
    click.echo("If you have ", nl=False)
    click.secho("API access ", nl=False, bold=True)
    click.echo("to udebug, you can now use icanc create testcases -u.")



import click
import sys

from .commands.create import create, scaffold
from .commands.init import init
from .commands.submit import submit
from .commands.test import test
from .commands.ci import ci
from .commands.udebug import udebug
from .commands.common.exception import IcancException

@click.group()
def icanc():
    pass

icanc.add_command(init)
icanc.add_command(create)
icanc.add_command(scaffold)
icanc.add_command(submit)
icanc.add_command(test)
icanc.add_command(ci)
icanc.add_command(udebug)

def main():
    try:
        icanc()
        sys.exit(0)
    except IcancException as e:
        click.secho(f"Error: {e.error}: {e.message}", err=True)
        if e.hint:
            click.secho(e.hint)
        sys.exit(-1)


if __name__ == '__main__':
    main()



import keyring

def set_auth(key, user, password):
    keyring.set_password(key, "user", user)
    keyring.set_password(key, "password", password)

def get_auth(key):
    return dict(
        user = keyring.get_password(key, "user"),
        password = keyring.get_password(key, "password"),
    )



import os
import tomllib

config = dict(
    author = None,
    email = None,
    repository = None,
    editor = "code.cmd",
    compiler = "gcc",
    udebug = None
)

def preload_rc():
    global config

    with open(os.path.join(os.getcwd(), "icancrc.toml"), "rb") as f:
        config_file = tomllib.load(f)
        config.update(**config_file)
